/**
 * @file account.c Galago Account message processor
 *
 * Copyright (C) 2004-2005 Christian Hammond.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA  02111-1307  USA
 */
#include "filter.h"
#include "meta-account.h"
#include "meta-presence.h"
#include "utils.h"

static void
_galagod_dbus_account_emit_connected(GalagoAccount *account)
{
	DBusMessage *message;

	galago_return_if_fail(account != NULL);

	message = dbus_message_new_signal(galago_object_get_dbus_path(account),
									  GALAGO_DBUS_ACCOUNT_INTERFACE,
									  "Connected");

	dbus_connection_send(galago_core_get_dbus_conn(), message, NULL);
	dbus_message_unref(message);
}

static void
_galagod_dbus_account_emit_disconnected(GalagoAccount *account)
{
	DBusMessage *message;

	galago_return_if_fail(account != NULL);

	message = dbus_message_new_signal(galago_object_get_dbus_path(account),
									  GALAGO_DBUS_ACCOUNT_INTERFACE,
									  "Disconnected");

	dbus_connection_send(galago_core_get_dbus_conn(), message, NULL);
	dbus_message_unref(message);
}

static void
_galagod_dbus_account_emit_display_name_changed(GalagoAccount *account)
{
	DBusMessage *message;
	DBusMessageIter iter;

	galago_return_if_fail(account != NULL);

	message = dbus_message_new_signal(galago_object_get_dbus_path(account),
									  GALAGO_DBUS_ACCOUNT_INTERFACE,
									  "DisplayNameChanged");

	dbus_message_iter_init(message, &iter);

	if (!galago_account_is_display_name_set(account))
		dbus_message_iter_append_nil(&iter);
	else
	{
		dbus_message_iter_append_string(&iter,
			galago_account_get_display_name(account));
	}

	dbus_connection_send(galago_core_get_dbus_conn(), message, NULL);
	dbus_message_unref(message);
}

static DBusHandlerResult
account_set_connected(DBusConnection *dbus_conn, DBusMessage *message,
					  GalagodCoCo *coco, GalagoAccount *account)
{
	const GalagoList *l;
	GalagodMetaAccount *meta_account;
	GalagoAccount *main_account;
	DBusMessageIter iter;
	galago_bool connected, old_connected;
	int connected_weight = 0;

	meta_account = GALAGOD_META_ACCOUNT(account);
	main_account = galagod_meta_account_get_account(meta_account);

	dbus_message_iter_init(message, &iter);
	connected = dbus_message_iter_get_boolean(&iter);

	if (connected == galago_account_is_connected(account))
		return DBUS_HANDLER_RESULT_HANDLED;

	galago_account_set_connected(account, connected);

	for (l = galagod_meta_account_get_accounts(meta_account);
		 l != NULL;
		 l = l->next)
	{
		GalagoAccount *temp_account = (GalagoAccount *)l->data;

		if (galago_account_is_connected(temp_account))
			connected_weight++;
		else
			connected_weight--;
	}

	old_connected = galago_account_is_connected(main_account);

	if (connected_weight > 0)
		connected = TRUE;
	else if (connected_weight < 0)
		connected = FALSE;
	else if (connected_weight == 0 && !connected)
		return DBUS_HANDLER_RESULT_HANDLED;

	if (old_connected != connected)
	{
		galago_account_set_connected(main_account, connected);

		if (connected)
			_galagod_dbus_account_emit_connected(main_account);
		else
			_galagod_dbus_account_emit_disconnected(main_account);
	}

	galagod_dbus_send_empty_reply(dbus_conn, message);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
account_add_contact(DBusConnection *dbus_conn, DBusMessage *message,
					GalagodCoCo *coco, GalagoAccount *account)
{
	GalagodMetaAccount *meta_account, *meta_contact;
	GalagoAccount *main_account, *main_contact;
	GalagoAccount *contact;
	DBusMessageIter iter;

	meta_account = GALAGOD_META_ACCOUNT(account);
	main_account = galagod_meta_account_get_account(meta_account);

	dbus_message_iter_init(message, &iter);
	contact = galago_dbus_message_iter_get_object(&iter, GALAGO_CLASS_ACCOUNT);

	meta_contact = GALAGOD_META_ACCOUNT(contact);
	main_contact = galagod_meta_account_get_account(meta_contact);

	galago_account_add_contact(main_account, main_contact);

	galagod_dbus_send_empty_reply(dbus_conn, message);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
account_remove_contact(DBusConnection *dbus_conn, DBusMessage *message,
					   GalagodCoCo *coco, GalagoAccount *account)
{
	GalagodMetaAccount *meta_account, *meta_contact;
	GalagoAccount *main_account, *main_contact;
	GalagoAccount *contact;
	DBusMessageIter iter;

	meta_account = GALAGOD_META_ACCOUNT(account);
	main_account = galagod_meta_account_get_account(meta_account);

	dbus_message_iter_init(message, &iter);
	contact = galago_dbus_message_iter_get_object(&iter,
												  GALAGO_CLASS_ACCOUNT);

	meta_contact = GALAGOD_META_ACCOUNT(contact);
	main_contact = galagod_meta_account_get_account(meta_contact);

	galago_account_remove_contact(main_account, main_contact);

	galagod_dbus_send_empty_reply(dbus_conn, message);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
account_get_contact(DBusConnection *dbus_conn, DBusMessage *message,
					GalagodCoCo *coco, GalagoAccount *account)
{
	GalagodMetaAccount *meta_account;
	GalagoAccount *main_account;
	GalagoAccount *contact;
	DBusMessageIter iter;
	DBusMessage *reply;
	char *username;

	meta_account = GALAGOD_META_ACCOUNT(account);
	main_account = galagod_meta_account_get_account(meta_account);

	dbus_message_iter_init(message, &iter);
	username = dbus_message_iter_get_string(&iter);

	contact = galago_account_get_contact(main_account, username, FALSE);

	dbus_free(username);

	reply = dbus_message_new_method_return(message);
	dbus_message_iter_init(reply, &iter);

	galago_dbus_message_iter_append_object(&iter, contact);

	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
account_get_contacts(DBusConnection *dbus_conn, DBusMessage *message,
					 GalagodCoCo *coco, GalagoAccount *account)
{
	GalagodMetaAccount *meta_account;
	GalagoAccount *main_account;
	const GalagoList *contacts;
	DBusMessageIter iter;
	DBusMessage *reply;

	meta_account = GALAGOD_META_ACCOUNT(account);
	main_account = galagod_meta_account_get_account(meta_account);

	contacts = galago_account_get_contacts(main_account, FALSE);

	reply = dbus_message_new_method_return(message);
	dbus_message_iter_init(reply, &iter);

	galago_dbus_message_iter_append_object_list(&iter, contacts);

	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
account_set_presence(DBusConnection *dbus_conn, DBusMessage *message,
					 GalagodCoCo *coco, GalagoAccount *account)
{
	GalagodMetaAccount *meta_account;
	GalagodMetaPresence *meta_presence = NULL;
	GalagoAccount *main_account;
	GalagoPresence *presence, *main_presence = NULL;
	DBusMessageIter iter;

	meta_account = GALAGOD_META_ACCOUNT(account);
	main_account = galagod_meta_account_get_account(meta_account);

	dbus_message_iter_init(message, &iter);
	presence = galago_dbus_message_iter_get_object(&iter,
												   GALAGO_CLASS_PRESENCE);

	if (galago_account_get_presence(main_account, FALSE) != NULL)
	{
		main_presence = galago_account_get_presence(main_account, FALSE);
		meta_presence = GALAGOD_META_PRESENCE(main_presence);
	}

	if (presence == NULL)
	{
		GalagoPresence *old_presence;

		old_presence = galago_account_get_presence(account, FALSE);

		if (meta_presence != NULL && old_presence != NULL)
		{
			galagod_meta_presence_remove_presence(meta_presence, old_presence);

			if (galagod_meta_presence_get_presences(meta_presence) == NULL)
				galagod_meta_presence_destroy(meta_presence);
		}
	}
	else
	{
		if (meta_presence == NULL)
			meta_presence = galagod_meta_presence_new(meta_account);

		galago_object_set_data(presence, "coco", coco);

		galagod_meta_presence_add_presence(meta_presence, presence);
	}

	galago_account_set_presence(account, presence);

	galagod_dbus_send_empty_reply(dbus_conn, message);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
account_get_presence(DBusConnection *dbus_conn, DBusMessage *message,
					 GalagodCoCo *coco, GalagoAccount *account)
{
	GalagodMetaAccount *meta_account;
	GalagoAccount *main_account;
	GalagoPresence *presence;
	DBusMessage *reply;
	DBusMessageIter iter;

	meta_account = GALAGOD_META_ACCOUNT(account);
	main_account = galagod_meta_account_get_account(meta_account);

	presence = galago_account_get_presence(main_account, FALSE);

	reply = dbus_message_new_method_return(message);
	dbus_message_iter_init(reply, &iter);

	galago_dbus_message_iter_append_object(&iter, presence);

	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
account_set_avatar(DBusConnection *dbus_conn, DBusMessage *message,
				   GalagodCoCo *coco, GalagoAccount *account)
{
	GalagodMetaAccount *meta_account;
	GalagoAccount *main_account;
	GalagoAvatar *avatar;
	DBusMessageIter iter;

	meta_account = GALAGOD_META_ACCOUNT(account);
	main_account = galagod_meta_account_get_account(meta_account);

	dbus_message_iter_init(message, &iter);
	avatar = galago_dbus_message_iter_get_object(&iter, GALAGO_CLASS_AVATAR);

	galago_account_set_avatar(main_account, avatar);

	galagod_dbus_send_empty_reply(dbus_conn, message);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
account_get_avatar(DBusConnection *dbus_conn, DBusMessage *message,
				   GalagodCoCo *coco, GalagoAccount *account)
{
	GalagodMetaAccount *meta_account;
	GalagoAccount *main_account;
	GalagoAvatar *avatar;
	DBusMessage *reply;
	DBusMessageIter iter;

	meta_account = GALAGOD_META_ACCOUNT(account);
	main_account = galagod_meta_account_get_account(meta_account);

	avatar = galago_account_get_avatar(main_account, FALSE);

	reply = dbus_message_new_method_return(message);
	dbus_message_iter_init(reply, &iter);

	galago_dbus_message_iter_append_object(&iter, avatar);

	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
account_set_display_name(DBusConnection *dbus_conn, DBusMessage *message,
						 GalagodCoCo *coco, GalagoAccount *account)
{
	GalagodMetaAccount *meta_account;
	GalagoAccount *main_account;
	char *display_name = NULL;
	DBusMessageIter iter;

	meta_account = GALAGOD_META_ACCOUNT(account);
	main_account = galagod_meta_account_get_account(meta_account);

	dbus_message_iter_init(message, &iter);

	if (dbus_message_iter_get_arg_type(&iter) != DBUS_TYPE_NIL)
		display_name = dbus_message_iter_get_string(&iter);

	galago_account_set_display_name(account,      display_name);
	galago_account_set_display_name(main_account, display_name);

	if (display_name != NULL)
		dbus_free(display_name);

	_galagod_dbus_account_emit_display_name_changed(account);

	galagod_dbus_send_empty_reply(dbus_conn, message);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
account_get_display_name(DBusConnection *dbus_conn, DBusMessage *message,
						 GalagodCoCo *coco, GalagoAccount *account)
{
	GalagodMetaAccount *meta_account;
	GalagoAccount *main_account;
	DBusMessage *reply;
	DBusMessageIter iter;

	meta_account = GALAGOD_META_ACCOUNT(account);
	main_account = galagod_meta_account_get_account(meta_account);

	reply = dbus_message_new_method_return(message);
	dbus_message_iter_init(reply, &iter);

	if (galago_account_is_display_name_set(main_account))
	{
		dbus_message_iter_append_string(&iter,
			galago_account_get_display_name(main_account));
	}
	else
		dbus_message_iter_append_nil(&iter);

	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
account_path_message_func(DBusConnection *dbus_conn, DBusMessage *message,
						  void *user_data)
{
	DBusHandlerResult result = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
	GalagodStateData *data = (GalagodStateData *)user_data;

	if (dbus_message_get_type(message) == DBUS_MESSAGE_TYPE_SIGNAL)
		return result;

	galago_log_misc("%s: %s\n",
					dbus_message_get_interface(message),
					dbus_message_get_member(message));

	galago_context_push(data->context);

	if (dbus_message_is_method_call(message,
									GALAGO_DBUS_ACCOUNT_INTERFACE,
									"SetConnected"))
	{
		result = account_set_connected(dbus_conn, message, data->coco,
									   data->u.account);
	}
	else if (dbus_message_is_method_call(message,
										 GALAGO_DBUS_ACCOUNT_INTERFACE,
										 "AddContact"))
	{
		result = account_add_contact(dbus_conn, message, data->coco,
									 data->u.account);
	}
	else if (dbus_message_is_method_call(message,
										 GALAGO_DBUS_ACCOUNT_INTERFACE,
										 "RemoveContact"))
	{
		result = account_remove_contact(dbus_conn, message, data->coco,
										data->u.account);
	}
	else if (dbus_message_is_method_call(message,
										 GALAGO_DBUS_ACCOUNT_INTERFACE,
										 "GetContact"))
	{
		result = account_get_contact(dbus_conn, message, data->coco,
									 data->u.account);
	}
	else if (dbus_message_is_method_call(message,
										 GALAGO_DBUS_ACCOUNT_INTERFACE,
										 "GetContacts"))
	{
		result = account_get_contacts(dbus_conn, message, data->coco,
									  data->u.account);
	}
	else if (dbus_message_is_method_call(message,
										 GALAGO_DBUS_ACCOUNT_INTERFACE,
										 "SetPresence"))
	{
		result = account_set_presence(dbus_conn, message, data->coco,
									  data->u.account);
	}
	else if (dbus_message_is_method_call(message,
										 GALAGO_DBUS_ACCOUNT_INTERFACE,
										 "GetPresence"))
	{
		result = account_get_presence(dbus_conn, message, data->coco,
									  data->u.account);
	}
	else if (dbus_message_is_method_call(message,
										 GALAGO_DBUS_ACCOUNT_INTERFACE,
										 "SetAvatar"))
	{
		result = account_set_avatar(dbus_conn, message, data->coco,
									data->u.account);
	}
	else if (dbus_message_is_method_call(message,
										 GALAGO_DBUS_ACCOUNT_INTERFACE,
										 "GetAvatar"))
	{
		result = account_get_avatar(dbus_conn, message, data->coco,
									data->u.account);
	}
	else if (dbus_message_is_method_call(message,
										 GALAGO_DBUS_ACCOUNT_INTERFACE,
										 "SetDisplayName"))
	{
		result = account_set_display_name(dbus_conn, message, data->coco,
										  data->u.account);
	}
	else if (dbus_message_is_method_call(message,
										 GALAGO_DBUS_ACCOUNT_INTERFACE,
										 "GetDisplayName"))
	{
		result = account_get_display_name(dbus_conn, message, data->coco,
										  data->u.account);
	}
	else
	{
		galago_log_warning("Unhandled message!\n");
		galago_log_misc("\tobj_path = '%s'\n",
						dbus_message_get_path(message));
		galago_log_misc("\tinterface = '%s'\n",
						dbus_message_get_interface(message));
		galago_log_misc("\tmethod = '%s'\n",
						dbus_message_get_member(message));
	}

	galago_context_pop();

	return result;
}

void
galagod_account_add_filter(GalagoAccount *account)
{
	galagod_filters_add(account, account_path_message_func);
}

void
galagod_account_remove_filter(GalagoAccount *account)
{
	galagod_filters_remove(account);
}
