/**
 * @file person.c Galago Person message processor
 *
 * Copyright (C) 2004-2005 Christian Hammond.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA  02111-1307  USA
 */
#include "filter.h"
#include "meta-person.h"
#include "utils.h"

static DBusHandlerResult
person_set_property(DBusConnection *dbus_conn, DBusMessage *message,
					GalagodCoCo *coco, GalagoPerson *person)
{
	GalagodMetaPerson *meta_person;
	GalagoPerson *main_person;
	char *name;
	DBusMessage *signal;
	DBusMessageIter iter, signal_iter;

	meta_person = GALAGOD_META_PERSON(person);
	main_person = galagod_meta_person_get_person(meta_person);

	dbus_message_iter_init(message, &iter);

	signal = dbus_message_new_signal(galago_object_get_dbus_path(main_person),
									 GALAGO_DBUS_PERSON_INTERFACE,
									 "PropertySet");
	dbus_message_iter_init(signal, &signal_iter);

	name = dbus_message_iter_get_string(&iter);
	dbus_message_iter_next(&iter);

	dbus_message_iter_append_string(&signal_iter, name);

	if (dbus_message_iter_get_arg_type(&iter) == DBUS_TYPE_STRING)
	{
		char *value = dbus_message_iter_get_string(&iter);
		galago_person_set_property_string(main_person, name, value);
		dbus_message_iter_append_string(&signal_iter, value);
		dbus_free(value);
	}
	else if (dbus_message_iter_get_arg_type(&iter) == DBUS_TYPE_BOOLEAN)
	{
		dbus_bool_t value = dbus_message_iter_get_boolean(&iter);
		galago_person_set_property_bool(main_person, name, value);
		dbus_message_iter_append_boolean(&signal_iter, value);
	}
	else if (dbus_message_iter_get_arg_type(&iter) == DBUS_TYPE_UINT32)
	{
		dbus_uint32_t value = dbus_message_iter_get_uint32(&iter);
		galago_person_set_property_uint32(main_person, name, value);
		dbus_message_iter_append_uint32(&signal_iter, value);
	}

	dbus_free(name);

	dbus_connection_send(dbus_conn, signal, NULL);
	dbus_message_unref(signal);

	galagod_dbus_send_empty_reply(dbus_conn, message);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
person_remove_property(DBusConnection *dbus_conn, DBusMessage *message,
					   GalagodCoCo *coco, GalagoPerson *person)
{
	GalagodMetaPerson *meta_person;
	GalagoPerson *main_person;
	char *name;
	DBusMessageIter iter;

	dbus_message_iter_init(message, &iter);
	name = dbus_message_iter_get_string(&iter);

	meta_person = GALAGOD_META_PERSON(person);
	main_person = galagod_meta_person_get_person(meta_person);

	galago_person_remove_property(main_person, name);

	dbus_free(name);

	galagod_dbus_send_empty_reply(dbus_conn, message);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
person_get_property(DBusConnection *dbus_conn, DBusMessage *message,
					GalagodCoCo *coco, GalagoPerson *person)
{
	GalagodMetaPerson *meta_person;
	GalagoPerson *main_person;
	const GalagoValue *value;
	char *name;
	DBusMessageIter iter;
	DBusMessage *reply;

	dbus_message_iter_init(message, &iter);
	name = dbus_message_iter_get_string(&iter);

	meta_person = GALAGOD_META_PERSON(person);
	main_person = galagod_meta_person_get_person(meta_person);

	value = galago_person_get_property(main_person, name);

	reply = dbus_message_new_method_return(message);
	dbus_message_iter_init(reply, &iter);

	if (value == NULL)
		dbus_message_iter_append_nil(&iter);
	else
	{
		GalagoType type = galago_value_get_type(value);

		if (type == GALAGO_TYPE_STRING)
		{
			dbus_message_iter_append_string(&iter,
				galago_value_get_string(value));
		}
		else if (type == GALAGO_TYPE_BOOLEAN)
		{
			dbus_message_iter_append_boolean(&iter,
				galago_value_get_boolean(value));
		}
		else if (type == GALAGO_TYPE_UINT)
		{
			dbus_message_iter_append_uint32(&iter,
				galago_value_get_uint(value));
		}
		else
		{
			/* This should never happen. */
			galago_log_fatal("Unknown property type %d for property %s\n",
							 type, name);
		}
	}

	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);

	dbus_free(name);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
person_get_account(DBusConnection *dbus_conn, DBusMessage *message,
				   GalagodCoCo *coco, GalagoPerson *person)
{
	GalagodMetaPerson *meta_person;
	GalagoPerson *main_person;
	GalagoAccount *account;
	GalagoService *service;
	char *username;
	DBusMessageIter iter;
	DBusMessage *reply;

	dbus_message_iter_init(message, &iter);
	service = galago_dbus_message_iter_get_object(&iter, GALAGO_CLASS_SERVICE);
	dbus_message_iter_next(&iter);
	username = dbus_message_iter_get_string(&iter);

	meta_person = GALAGOD_META_PERSON(person);
	main_person = galagod_meta_person_get_person(meta_person);

	account = galago_person_get_account(main_person, service, username, FALSE);

	dbus_free(username);

	reply = dbus_message_new_method_return(message);
	dbus_message_iter_init(reply, &iter);

	galago_dbus_message_iter_append_object(&iter, account);

	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
person_get_accounts(DBusConnection *dbus_conn, DBusMessage *message,
					GalagodCoCo *coco, GalagoPerson *person)
{
	GalagodMetaPerson *meta_person;
	GalagoPerson *main_person;
	const GalagoList *list;
	DBusMessageIter iter;
	DBusMessage *reply;

	meta_person = GALAGOD_META_PERSON(person);
	main_person = galagod_meta_person_get_person(meta_person);

	list = galago_person_get_accounts(main_person, FALSE);

	reply = dbus_message_new_method_return(message);
	dbus_message_iter_init(reply, &iter);

	galago_dbus_message_iter_append_object_list(&iter, list);

	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
person_set_photo(DBusConnection *dbus_conn, DBusMessage *message,
				 GalagodCoCo *coco, GalagoPerson *person)
{
	GalagodMetaPerson *meta_person;
	GalagoPerson *main_person;
	GalagoPhoto *photo;
	DBusMessageIter iter;

	meta_person = GALAGOD_META_PERSON(person);
	main_person = galagod_meta_person_get_person(meta_person);

	dbus_message_iter_init(message, &iter);
	photo = galago_dbus_message_iter_get_object(&iter, GALAGO_CLASS_PHOTO);

	galago_person_set_photo(main_person, photo);

	galagod_dbus_send_empty_reply(dbus_conn, message);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
person_get_photo(DBusConnection *dbus_conn, DBusMessage *message,
				 GalagodCoCo *coco, GalagoPerson *person)
{
	GalagodMetaPerson *meta_person;
	GalagoPerson *main_person;
	GalagoPhoto *photo;
	DBusMessage *reply;
	DBusMessageIter iter;

	meta_person = GALAGOD_META_PERSON(person);
	main_person = galagod_meta_person_get_person(meta_person);

	photo = galago_person_get_photo(main_person, FALSE);

	reply = dbus_message_new_method_return(message);
	dbus_message_iter_init(reply, &iter);

	galago_dbus_message_iter_append_object(&iter, photo);

	dbus_connection_send(dbus_conn, reply, NULL);
	dbus_message_unref(reply);

	return DBUS_HANDLER_RESULT_HANDLED;
}

static DBusHandlerResult
person_path_message_func(DBusConnection *dbus_conn, DBusMessage *message,
						 void *user_data)
{
	DBusHandlerResult result = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;

	GalagodStateData *data = (GalagodStateData *)user_data;

	if (dbus_message_get_type(message) == DBUS_MESSAGE_TYPE_SIGNAL)
		return result;

	galago_log_misc("%s: %s\n",
					dbus_message_get_interface(message),
					dbus_message_get_member(message));

	galago_context_push(data->context);

	if (dbus_message_is_method_call(message,
									GALAGO_DBUS_PERSON_INTERFACE,
									"SetProperty"))
	{
		result = person_set_property(dbus_conn, message, data->coco,
									 data->u.person);
	}
	else if (dbus_message_is_method_call(message,
										 GALAGO_DBUS_PERSON_INTERFACE,
										 "RemoveProperty"))
	{
		result = person_remove_property(dbus_conn, message, data->coco,
										data->u.person);
	}
	else if (dbus_message_is_method_call(message,
										 GALAGO_DBUS_PERSON_INTERFACE,
										 "GetProperty"))
	{
		result = person_get_property(dbus_conn, message, data->coco,
									 data->u.person);
	}
	else if (dbus_message_is_method_call(message,
										 GALAGO_DBUS_PERSON_INTERFACE,
										 "GetAccount"))
	{
		result = person_get_account(dbus_conn, message, data->coco,
									data->u.person);
	}
	else if (dbus_message_is_method_call(message,
										 GALAGO_DBUS_PERSON_INTERFACE,
										 "GetAccounts"))
	{
		result = person_get_accounts(dbus_conn, message, data->coco,
									 data->u.person);
	}
	else if (dbus_message_is_method_call(message,
										 GALAGO_DBUS_PERSON_INTERFACE,
										 "SetPhoto"))
	{
		result = person_set_photo(dbus_conn, message, data->coco,
								  data->u.person);
	}
	else if (dbus_message_is_method_call(message,
										 GALAGO_DBUS_PERSON_INTERFACE,
										 "GetPhoto"))
	{
		result = person_get_photo(dbus_conn, message, data->coco,
								  data->u.person);
	}
	else
	{
		galago_log_warning("Unhandled message!\n");
		galago_log_misc("\tobj_path = '%s'\n",
						dbus_message_get_path(message));
		galago_log_misc("\tinterface = '%s'\n",
						dbus_message_get_interface(message));
		galago_log_misc("\tmethod = '%s'\n",
						dbus_message_get_member(message));
	}

	galago_context_pop();

	return result;
}

void
galagod_person_add_filter(GalagoPerson *person)
{
	galagod_filters_add(person, person_path_message_func);
}

void
galagod_person_remove_filter(GalagoPerson *person)
{
	galagod_filters_remove(person);
}
